try:
    from importlib.metadata import PackageNotFoundError, metadata  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, metadata  # type: ignore

from .resource_id import ResourceId

__all__ = ["ResourceId"]


try:
    __version__: str = metadata(__name__)["version"]
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = ""



"""ResourceId implements base62-encoded identifiers, suitable for URLs and URIs."""

from typing import Any, Dict, Protocol, Union, runtime_checkable
from uuid import UUID

try:
    from typing import TypeAlias
except ImportError:  # pragma: no cover
    from typing_extensions import TypeAlias


try:  # pragma: no cover
    from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
    from pydantic.json_schema import JsonSchemaValue
    from pydantic_core import CoreSchema, core_schema

    pydantic2 = True
except ImportError:  # pragma: no cover
    pydantic2 = False

__all__ = ["ResourceId"]

ALPHABET = tuple("0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")
DECODE_MAP = {x: idx for idx, x in enumerate(ALPHABET)}


@runtime_checkable
class Base62Encodable(Protocol):
    def __int__(self) -> int:  # pragma: no cover
        ...


ResourceIdValue: TypeAlias = Union[str, Base62Encodable]


def b62encode(value: Base62Encodable):
    """Encode anything that can be converted to a non-negative int.  This includes uuid.UUID objects."""
    value = int(value)
    if 0 > value:
        raise ValueError("value must convert to a non-negative integer.")
    x = value
    b62_repr: str = ""
    while x:
        x, b62_repr = (
            x // 62,
            ALPHABET[x % 62] + b62_repr,
        )
    return b62_repr if b62_repr else "0"


def b62decode(value: str):
    """Decode a base62-encoded str.  Returns int.  Raises ValueError if value is invalid."""

    sgn, value = (-1, value[1:]) if value[0] == "-" else (1, value)
    x = 0
    for digit in value:
        try:
            x = x * 62 + DECODE_MAP[digit]
        except KeyError as exc:
            raise ValueError(f"Invalid base62 value '{value}'.") from exc
    return sgn * x


class ResourceId:
    __slots__ = ["value"]

    def __init__(self, value: ResourceIdValue):
        self.value = self._to_int(value)

    @property
    def uuid(self) -> UUID:
        return UUID(int=self.value)

    def __repr__(self) -> str:
        return b62encode(self.value)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self.value == other.value
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.value)

    def __int__(self):
        return self.value

    @staticmethod
    def _to_int(value: ResourceIdValue):
        # to be replaced with a match statement someday.
        if isinstance(value, str):
            return b62decode(value)
        elif isinstance(value, Base62Encodable):  # type: ignore
            int_value = int(value)
            if int_value < 0:
                raise ValueError("value must be non-negative.")
            return int_value
        else:
            raise TypeError("value must be str or Base62Encodable.")

    @classmethod
    def _json_schema(cls):
        return {
            "title": cls.__name__,
            "description": "An opaque identifier.",
            "type": "string",
            "format": "resource-id",
        }

    if pydantic2:
        # this is a little too dynamic for pyright, so we disable a few type warnings.

        @classmethod
        def __get_pydantic_core_schema__(  # type: ignore
            cls, source_type: Any, handler: GetCoreSchemaHandler  # type: ignore
        ) -> CoreSchema:  # type: ignore
            return core_schema.no_info_plain_validator_function(cls, serialization=core_schema.plain_serializer_function_ser_schema(str, return_schema=core_schema.str_schema()))  # type: ignore

        @classmethod
        def __get_pydantic_json_schema__(
            cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler  # type: ignore
        ) -> JsonSchemaValue:
            return cls._json_schema()

    else:

        @classmethod
        def __get_validators__(cls):
            # for pydantic 1
            yield cls.validate

        @classmethod
        def __modify_schema__(cls, field_schema: Dict[str, Any]):
            # __modify_schema__ should mutate the dict it receives in place,
            # the returned value will be ignored
            # __get_pydantic_json_schema__ replaces __modify_schema__ in pydantic 2.
            field_schema.update(cls._json_schema())

        @classmethod
        def validate(cls, value: ResourceIdValue):
            return cls(value)



from typing import Union
from uuid import UUID

import jsonschema
import pydantic
import pytest

from resource_id.resource_id import Base62Encodable, ResourceId, b62decode, b62encode

pydantic_major_version = int(pydantic.VERSION.split(".")[0])


@pytest.mark.parametrize("src, expected", [(1, "1"), (62, "10")])
def test_b62encode(src: Base62Encodable, expected: str):
    assert b62encode(src) == expected


def test_b62encode_fail():
    with pytest.raises(ValueError):
        b62encode(-1)


@pytest.mark.parametrize("src, expected", [("1", 1), ("11", 63)])
def test_b62decode(src: str, expected: int):
    assert b62decode(src) == expected


def test_b62decode_fail():
    with pytest.raises(ValueError):
        b62decode("*")


@pytest.mark.parametrize("arg, value", [(1, 1), ("11", 63), (UUID(int=1728), 1728)])
def test_init(arg: Union[str, Base62Encodable], value: int):
    assert ResourceId(arg).value == value


def test_init_bad_arg_value():
    with pytest.raises(ValueError):
        ResourceId(-1)


def test_init_bad_arg_type():
    with pytest.raises(TypeError):
        ResourceId({})  # type: ignore


def test_uuid():
    value = UUID(int=123)
    assert ResourceId(value).uuid == value


def test_repr():
    arg = 1
    assert repr(ResourceId(arg)) == "1"


def test_eq():
    assert ResourceId(1) == ResourceId(1)


def test_not_eq():
    assert ResourceId(1) != 1


def test_hash():
    assert hash(ResourceId(1)) == hash(1)


def test_int():
    value = 2
    assert int(ResourceId(value)) == value


def test___get_validators__():
    if pydantic_major_version == 1:
        assert next(ResourceId.__get_validators__()) == ResourceId.validate  # type: ignore


def test_validate():
    if pydantic_major_version == 1:
        assert ResourceId.validate("test") == ResourceId("test")
    else:
        t = pydantic.TypeAdapter(ResourceId)
        assert t.validate_json('"test"') == ResourceId("test")


def test_modify_schema():
    if pydantic_major_version == 1:
        __modify_schema__ = ResourceId.__origin__.__modify_schema__ if hasattr(ResourceId, "__origin__") else ResourceId.__modify_schema__  # type: ignore
        schema = {}
        __modify_schema__(schema)
        assert schema == ResourceId._json_schema()  # type: ignore


def test__json_schema():
    if pydantic_major_version == 1:
        assert isinstance(ResourceId._json_schema(), dict)  # type: ignore


@pytest.mark.skipif(
    not hasattr(pydantic, "TypeAdapter"),
    reason="Serialization is only implemented for pydantic 2.",
)
def test_serialization_pydantic2():
    V = pydantic.TypeAdapter(ResourceId)
    id = ResourceId("test")
    assert V.dump_json(id) == b'"test"'


@pytest.mark.skipif(
    pydantic_major_version != 2,
    reason="model_json_schema is defined only for pydantic 2.",
)
def test_model_json_schema():
    class Model(pydantic.BaseModel):
        id: ResourceId

    m = Model(id=ResourceId("test"))
    assert m.model_json_schema(mode="serialization")

def test_json_schema():
    assert jsonschema.validate('test', ResourceId._json_schema()) is None # pyright: ignore[reportPrivateUsage]
